/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.binaryfile;

import com.google.protobuf.ByteString;
import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.Context;
import io.grpc.stub.CallStreamObserver;
import lombok.Builder;
import lombok.extern.log4j.Log4j2;
import org.apache.commons.io.IOUtils;
import org.springframework.core.task.TaskExecutor;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.function.Function;

@Log4j2
public class GrpcBinaryFileServerDownloader<T> {

	private static final int CHUNK_SIZE = 255 * 1024;

	private final CallStreamObserver<T> callObserver;
	private final Function<ByteString, T> chunkBuilder;
	private final Consumer<OutputStream> downloadConsumer;
	private final TaskExecutor taskExecutor;

	private final byte[] buffer = new byte[GrpcBinaryFileServerDownloader.CHUNK_SIZE];
	private final AtomicBoolean started = new AtomicBoolean(false);
	private final AtomicBoolean downloadInProgress = new AtomicBoolean(false);

	private PipedInputStream inputStream;
	private PipedOutputStream outputStream;

	@Builder
	public GrpcBinaryFileServerDownloader(CallStreamObserver<T> callObserver, Function<ByteString, T> chunkBuilder,
			Consumer<OutputStream> downloadConsumer, TaskExecutor taskExecutor) {
		this.callObserver = callObserver;
		this.chunkBuilder = chunkBuilder;
		this.downloadConsumer = downloadConsumer;
		this.taskExecutor = taskExecutor;
	}

	public void start() {
		if (isNotStarted()) {
			doStart();
		}
	}

	boolean isNotStarted() {
		return started.compareAndSet(false, true);
	}

	void doStart() {
		handleSafety(this::setupStreams);
		taskExecutor.execute(Context.current().wrap(this::startDownload));
		callObserver.setOnReadyHandler(this::onReadyHandler);
	}

	void setupStreams() throws IOException {
		outputStream = new PipedOutputStream();
		inputStream = new PipedInputStream(GrpcBinaryFileServerDownloader.CHUNK_SIZE);
		outputStream.connect(inputStream);
	}

	void startDownload() {
		handleSafety(this::doDownload);
	}

	void doDownload() throws IOException {
		downloadInProgress.set(true);
		downloadConsumer.accept(outputStream);
		downloadInProgress.set(false);
		outputStream.close();
	}

	synchronized void onReadyHandler() {
		if (callObserver.isReady()) {
			sendChunks();
		}
	}

	void sendChunks() {
		handleSafety(this::doSendChunks);
	}

	void doSendChunks() throws IOException {
		int bytesRead;
		while (callObserver.isReady() && (bytesRead = inputStream.read(buffer)) != -1) {
			callObserver.onNext(chunkBuilder.apply(ByteString.copyFrom(buffer, 0, bytesRead)));
		}
		if (!downloadInProgress.get()) {
			inputStream.close();
			callObserver.onCompleted();
		}
	}

	void handleSafety(ExceptionalRunnable runnable) {
		try {
			runnable.run();
		} catch (Exception e) {
			IOUtils.closeQuietly(inputStream, e1 -> LOG.error("InputStream cannot be closed.", e1));
			IOUtils.closeQuietly(outputStream, e1 -> LOG.error("OutputStream cannot be closed.", e1));
			throw new TechnicalException("Error occurred during downloading file content download.", e);
		}
	}
}